"""Full view tests for django_ca_cmc."""

from http import HTTPStatus

import asn1crypto.cms
import asn1crypto.x509
import pytest
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, ed448, ed25519, rsa
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.x509.oid import CertificatePoliciesOID, ExtensionOID, NameOID
from django.test import Client
from django.urls import reverse
from django_ca.models import Certificate, CertificateAuthority
from pytest_django.fixtures import SettingsWrapper

from django_ca_cmc.models import CMCClient
from tests.utils import create_cmc_request, create_tagged_csr_message, load_file


@pytest.mark.usefixtures("pre_created_client")
def test_pre_created_csr(client: Client, ca: CertificateAuthority) -> None:
    """Test valid, pre-created request payloads."""
    public_key = ca.pub.loaded.public_key()
    if isinstance(public_key, ec.EllipticCurvePublicKey) and isinstance(
        public_key.curve, ec.SECT571K1
    ):
        pytest.xfail("This curve is known to not be supported at the moment.")

    assert ca.certificate_set.all().count() == 0  # just assert initial state

    decoded = load_file("cmc_with_csr")
    url_path = reverse("django_ca_cmc:cmc", kwargs={"serial": ca.serial})
    response = client.post(url_path, data=decoded, content_type="application/pkcs7-mime")
    assert response.status_code == HTTPStatus.OK, response.content

    certificate: Certificate = ca.certificate_set.get()

    info = asn1crypto.cms.ContentInfo.load(response.content)

    # Check content type
    content_type = info["content_type"]
    assert isinstance(content_type, asn1crypto.cms.ContentType)
    assert content_type.dotted == "1.2.840.113549.1.7.2"  # signed_data

    # Check content
    content = info["content"]
    assert isinstance(content, asn1crypto.cms.SignedData)
    assert content.native["version"] == "v2"

    # Check certificates in the response
    certificates = [cert.chosen for cert in content["certificates"]]
    assert isinstance(content["certificates"], asn1crypto.cms.CertificateSet)
    assert all(isinstance(cert, asn1crypto.x509.Certificate) for cert in certificates)
    assert set(cert.dump() for cert in certificates) == set([ca.pub.der, certificate.pub.der])

    signer_infos = content["signer_infos"]
    assert isinstance(signer_infos, asn1crypto.cms.SignerInfos)
    assert len(signer_infos) == 1

    signer_info = signer_infos[0]
    signature = signer_info["signature"].contents
    signed_data = signer_info["signed_attrs"].retag(17).dump()

    # Verify signature
    if isinstance(public_key, rsa.RSAPublicKey):
        algorithm = ca.algorithm
        assert algorithm is not None  # just to make mypy happy
        public_key.verify(signature, signed_data, algorithm=algorithm, padding=PKCS1v15())
    elif isinstance(public_key, ec.EllipticCurvePublicKey):
        public_key.verify(signature, signed_data, ec.ECDSA(hashes.SHA512()))
    elif isinstance(public_key, ed448.Ed448PublicKey | ed25519.Ed25519PublicKey):
        public_key.verify(signature, signed_data)
    else:
        raise ValueError("New key type?")


@pytest.mark.usefixtures("pre_created_client")
def test_pre_created_crmf(client: Client, rsa_2048_sha256_ca: CertificateAuthority) -> None:
    """Test valid, pre-created request payloads."""
    decoded = load_file("cmc_with_crmf")
    url_path = reverse("django_ca_cmc:cmc", kwargs={"serial": rsa_2048_sha256_ca.serial})
    response = client.post(url_path, data=decoded, content_type="application/pkcs7-mime")
    assert response.status_code == HTTPStatus.OK, response.content


def test_with_copied_extensions(
    settings: SettingsWrapper,
    cmc_client_private_key: ec.EllipticCurvePrivateKey,
    cmc_client: CMCClient,
    client: Client,
    rsa_private_key_2048: rsa.RSAPrivateKey,
    rsa_2048_sha256_ca: CertificateAuthority,
) -> None:
    """Test copying extensions."""
    csr_builder = (
        x509.CertificateSigningRequestBuilder()
        .subject_name(x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, "example.com")]))
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName("example.com")]),
            critical=False,
        )
        .add_extension(
            x509.CertificatePolicies(
                [x509.PolicyInformation(CertificatePoliciesOID.ANY_POLICY, None)]
            ),
            critical=False,
        )
        .add_extension(
            x509.UnrecognizedExtension(oid=x509.ObjectIdentifier("1.2.3"), value=b"123"),
            critical=False,
        )
        # This extension is blacklisted and will not be copied.
        .add_extension(x509.BasicConstraints(ca=True, path_length=0), critical=True)
    )
    csr = csr_builder.sign(rsa_private_key_2048, hashes.SHA256())

    cmc_client.copy_extensions = True
    cmc_client.save()
    settings.CA_CMC_COPY_UNRECOGNIZED_CSR_EXTENSIONS = True

    tcr = create_tagged_csr_message(csr)
    decoded = create_cmc_request(cmc_client_private_key, cmc_client.certificate.loaded, tcr)

    url_path = reverse("django_ca_cmc:cmc", kwargs={"serial": rsa_2048_sha256_ca.serial})
    response = client.post(url_path, data=decoded, content_type="application/pkcs7-mime")
    assert response.status_code == HTTPStatus.OK, response.content

    cert: Certificate = Certificate.objects.get()

    # CertificatePolicies is copied
    assert cert.extensions[ExtensionOID.CERTIFICATE_POLICIES] == x509.Extension(
        oid=ExtensionOID.CERTIFICATE_POLICIES,
        critical=False,
        value=x509.CertificatePolicies(
            [x509.PolicyInformation(CertificatePoliciesOID.ANY_POLICY, None)]
        ),
    )

    # UnrecognizedExtension is also copied
    assert cert.extensions[x509.ObjectIdentifier("1.2.3")] == x509.Extension(
        critical=False,
        oid=x509.ObjectIdentifier("1.2.3"),
        value=x509.UnrecognizedExtension(oid=x509.ObjectIdentifier("1.2.3"), value=b"123"),
    )

    # BasicConstraints is in the blacklist, so it's not copied
    assert cert.extensions[ExtensionOID.BASIC_CONSTRAINTS] == x509.Extension(
        critical=True,
        oid=ExtensionOID.BASIC_CONSTRAINTS,
        value=x509.BasicConstraints(ca=False, path_length=None),
    )


@pytest.mark.usefixtures("pre_created_client")
def test_pre_created_request_with_invalid_signature(
    client: Client, rsa_2048_sha256_ca: CertificateAuthority
) -> None:
    """Test pre-created request with an invalid signature."""
    decoded = load_file("cmc_with_invalid_signature")
    url_path = reverse("django_ca_cmc:cmc", kwargs={"serial": rsa_2048_sha256_ca.serial})
    response = client.post(url_path, data=decoded, content_type="application/pkcs7-mime")
    assert response.status_code == HTTPStatus.FORBIDDEN
    assert response.content == b"invalid signature"


def test_invalid_content_type(client: Client, rsa_2048_sha256_ca: CertificateAuthority) -> None:
    """Test sending an invalid content type."""
    url_path = reverse("django_ca_cmc:cmc", kwargs={"serial": rsa_2048_sha256_ca.serial})
    response = client.post(url_path, data=b"", content_type="text/plain")
    assert response.status_code == HTTPStatus.BAD_REQUEST
    assert response.content == b"invalid content type"



"""Utility functions for unit tests."""

import hashlib
from datetime import UTC, datetime
from importlib import resources

import asn1crypto.cms
import asn1crypto.csr
import asn1crypto.util
import asn1crypto.x509
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import Encoding
from python_cmc import cmc

from tests import files


def load_file(path: str) -> bytes:
    """Utility function to load a file from ``tests.files``."""
    full_path = resources.files(files) / path
    with full_path.open("rb") as stream:
        return stream.read()


def create_tagged_csr_message(csr: x509.CertificateSigningRequest) -> bytes:
    """Create a tagged CSR message for use in a CMC request."""
    encoded_csr = csr.public_bytes(Encoding.DER)

    tcr = cmc.TaggedCertificationRequest()
    tcr["bodyPartID"] = 1185658366
    tcr["certificationRequest"] = asn1crypto.csr.CertificationRequest.load(encoded_csr)
    tagged_request = cmc.TaggedRequest(name="tcr", value=tcr)

    controls = cmc.Controls()
    controls.append(
        asn1crypto.util.OrderedDict(
            {
                "bodyPartID": 340570457,
                "attrType": "id-cmc-senderNonce",
                "attrValues": [
                    b'S\xc3f\xa5O/\x15\xb6\xfe\x07"\x04\xfe\xba\xf2\x94H\xf4\x04\xac\xedv\x96\x95'
                    b"\xe7Y\xcf\xcc]T\xe0d\x80\x9a\xd8\x87\xdejb\xb1\xef.\x90\xda\x96#O\x90\xb4Z"
                    b"\xec~\xb2\xad\xc4Z\xcb\xb5\xbe\n\x8c\x9a\xa8\xcd\x04\xf01Y\xa4\xf0\ng\x03>"
                    b"\xa5\x97\xa9\x1f\x95\x15\x07\x84\x9bF\x90\x12\xb0\x15+&\x80F\xeb\x17xX\x17"
                    b"\x04l\xf6\xf2\xc4\xca\x89\\\xb4\xf2\x0b#v{\xdd_@\x15\xfe\x99\x11\xf10o\xb9"
                    b"\xf2\r\xf8`\x89\x91"
                ],
            }
        )
    )
    controls.append(
        asn1crypto.util.OrderedDict(
            {"bodyPartID": 937138838, "attrType": "id-cmc-regInfo", "attrValues": [b"pkcs10"]}
        ),
    )

    pki_data = cmc.PKIData()
    pki_data["reqSequence"] = cmc.TaggedRequests([tagged_request])
    pki_data["controlSequence"] = controls
    pki_data["cmsSequence"] = []
    pki_data["otherMsgSequence"] = []

    return pki_data.dump()  # type: ignore[no-any-return]


def create_cmc_request(
    private_key: ec.EllipticCurvePrivateKey, certificate: x509.Certificate, data: bytes
) -> bytes:
    """Create a signed CMC request with the given data."""
    certificate_der = certificate.public_bytes(Encoding.DER)
    asn1crypto_certificate = asn1crypto.x509.Certificate.load(certificate_der)

    # Start creating SignedData instance (equals content_info['content'])
    signed_data = asn1crypto.cms.SignedData()
    signed_data["version"] = asn1crypto.cms.CMSVersion("v3")  # tested
    signed_data["encap_content_info"] = asn1crypto.util.OrderedDict(
        [("content_type", "1.3.6.1.5.5.7.12.2"), ("content", data)]
    )
    signed_data["digest_algorithms"] = [  # tested
        asn1crypto.util.OrderedDict([("algorithm", "sha256"), ("parameters", None)])
    ]
    signed_data["certificates"] = asn1crypto.cms.CertificateSet([asn1crypto_certificate])

    # Setting signer info section
    signer_info = asn1crypto.cms.SignerInfo()
    signer_info["version"] = "v1"  # tested
    signer_info["digest_algorithm"] = asn1crypto.util.OrderedDict(  # tested
        [("algorithm", "sha256"), ("parameters", None)]
    )
    signer_info["signature_algorithm"] = asn1crypto.util.OrderedDict(  # tested
        [("algorithm", "sha256_ecdsa"), ("parameters", None)]
    )

    # Finding subject_key_identifier from certificate (asn1crypto.x509 object)
    key_id = asn1crypto_certificate.key_identifier_value.native
    signer_info["sid"] = asn1crypto.cms.SignerIdentifier({"subject_key_identifier": key_id})

    # Adding CMS attributes to singer_infos
    cms_attributes = asn1crypto.cms.CMSAttributes()
    cms_attributes.append(
        asn1crypto.cms.CMSAttribute(
            {
                "type": asn1crypto.cms.CMSAttributeType("content_type"),
                "values": asn1crypto.cms.SetOfContentType(
                    [asn1crypto.cms.ContentType("1.3.6.1.5.5.7.12.2")]
                ),
            }
        )
    )
    cms_attributes.append(
        asn1crypto.cms.CMSAttribute(
            {
                "type": asn1crypto.cms.CMSAttributeType("signing_time"),
                "values": asn1crypto.cms.SetOfTime([asn1crypto.core.UTCTime(datetime.now(UTC))]),
            }
        )
    )
    cms_attributes.append(
        asn1crypto.cms.CMSAttribute(
            {
                "type": asn1crypto.cms.CMSAttributeType("cms_algorithm_protection"),
                "values": asn1crypto.cms.SetOfCMSAlgorithmProtection(
                    [
                        asn1crypto.cms.CMSAlgorithmProtection(
                            {
                                "digest_algorithm": asn1crypto.util.OrderedDict(  # tested
                                    [("algorithm", "sha256"), ("parameters", None)]
                                ),
                                "signature_algorithm": asn1crypto.util.OrderedDict(  # tested
                                    [("algorithm", "sha256_ecdsa"), ("parameters", None)]
                                ),
                            }
                        )
                    ]
                ),
            }
        )
    )

    # compute and set message digest
    message_digest = hashlib.sha256(signed_data["encap_content_info"]["content"].contents).digest()
    cms_attributes.append(
        asn1crypto.cms.CMSAttribute(
            {
                "type": asn1crypto.cms.CMSAttributeType("message_digest"),
                "values": asn1crypto.cms.SetOfOctetString([message_digest]),
            }
        )
    )
    signer_info["signed_attrs"] = cms_attributes

    # Creating a signature using a private key object from pkcs11
    sign_data = signer_info["signed_attrs"].retag(17).dump()
    signer_info["signature"] = private_key.sign(sign_data, ec.ECDSA(hashes.SHA256()))

    # Add signer infos to signed data
    signed_data["signer_infos"] = asn1crypto.cms.SignerInfos({signer_info})

    # Create top-level ContentInfo object.
    content_info = asn1crypto.cms.ContentInfo()
    content_info["content_type"] = "signed_data"  # NOTE: must be BEFORE content
    content_info["content"] = signed_data

    return content_info.dump()  # type: ignore[no-any-return]



"""Utility functions, mostly for converting from/to cryptography/asn1crypto."""

from typing import cast

import asn1crypto.algos
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, ed448, ed25519, rsa
from django_ca.typehints import AllowedHashTypes


def _get_ec_signed_digest_algorithm(
    curve: ec.EllipticCurve,
) -> asn1crypto.algos.SignedDigestAlgorithm:
    # Helper function to get the digest algorithm for EC keys.
    if isinstance(curve, ec.SECP256R1):
        return asn1crypto.algos.SignedDigestAlgorithmId("sha256_ecdsa")
    elif isinstance(curve, ec.SECP384R1):
        return asn1crypto.algos.SignedDigestAlgorithmId("sha384_ecdsa")
    elif isinstance(curve, ec.SECP521R1):
        return asn1crypto.algos.SignedDigestAlgorithmId("sha512_ecdsa")
    else:
        # TODO: Support other curves
        raise ValueError(f"{curve.name}: Elliptic curve not supported.")


def get_signed_digest_algorithm(
    certificate: x509.Certificate,
) -> asn1crypto.algos.SignedDigestAlgorithm:
    """Get the ``asn1crypto.algos.SignedDigestAlgorithm`` of the given certificate."""
    # equivalent to signed_digest_algo() in
    # https://github.com/SUNET/pkcs11_ca/blob/main/tests/lib.py#L16
    algo = asn1crypto.algos.SignedDigestAlgorithm()
    if not isinstance(certificate, x509.Certificate):
        raise TypeError(f"{certificate}: Must be of type cryptography.x509.Certificate.")

    public_key = certificate.public_key()
    if isinstance(public_key, ed25519.Ed25519PublicKey):
        algo["algorithm"] = asn1crypto.algos.SignedDigestAlgorithmId("ed25519")
    elif isinstance(public_key, ed448.Ed448PublicKey):
        algo["algorithm"] = asn1crypto.algos.SignedDigestAlgorithmId("ed448")
    elif isinstance(public_key, ec.EllipticCurvePublicKey):
        algo["algorithm"] = _get_ec_signed_digest_algorithm(public_key.curve)
    elif isinstance(public_key, rsa.RSAPublicKey):
        algorithm = cast(AllowedHashTypes, certificate.signature_hash_algorithm)
        if isinstance(algorithm, hashes.SHA224):
            algo["algorithm"] = asn1crypto.algos.SignedDigestAlgorithmId("sha224_rsa")
        elif isinstance(algorithm, hashes.SHA256):
            algo["algorithm"] = asn1crypto.algos.SignedDigestAlgorithmId("sha256_rsa")
        elif isinstance(algorithm, hashes.SHA384):
            algo["algorithm"] = asn1crypto.algos.SignedDigestAlgorithmId("sha384_rsa")
        elif isinstance(algorithm, hashes.SHA512):
            algo["algorithm"] = asn1crypto.algos.SignedDigestAlgorithmId("sha512_rsa")
        else:
            raise ValueError(f"{algorithm.name}: Signature hash algorithm not supported.")
    else:
        # TODO: Add support for DSA
        raise ValueError(f"{public_key}: Public key type not supported.")

    return algo



"""
pytest configuration and fixtures.

.. seealso:: https://docs.pytest.org/en/stable/reference/fixtures.html
"""

from collections.abc import Callable
from datetime import UTC, datetime, timedelta
from typing import cast

import pytest
from _pytest.fixtures import SubRequest
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import dsa, ec, ed448, ed25519, rsa
from cryptography.hazmat.primitives.asymmetric.types import CertificateIssuerPrivateKeyTypes
from cryptography.x509.oid import NameOID
from django_ca.key_backends.db.models import DBStorePrivateKeyOptions
from django_ca.models import CertificateAuthority
from django_ca.typehints import AllowedHashTypes

from django_ca_cmc.models import CMCClient
from tests.utils import load_file

CMC_CLIENT_ONE_PEM = load_file("cmc_client_1.pem")

RSA_KEY_SIZES = (2048, 4096)
RSA_ALGORITHMS = (hashes.SHA224(), hashes.SHA256(), hashes.SHA384(), hashes.SHA512())
ELLIPTIC_CURVES = [ec.SECP521R1(), ec.SECP256R1(), ec.SECP384R1(), ec.SECT571K1()]


def _sign(
    private_key: CertificateIssuerPrivateKeyTypes,
    common_name: str,
    algorithm: AllowedHashTypes | None = None,
    ca: bool = True,
) -> x509.Certificate:
    subject = issuer = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, common_name)])

    now = datetime.now(UTC)
    serial = x509.random_serial_number()
    public_key = private_key.public_key()

    builder = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(public_key)
        .serial_number(serial)
        .not_valid_before(now)
        .not_valid_after(now + timedelta(days=10))
        # Add necessary extensions:
        .add_extension(x509.AuthorityKeyIdentifier.from_issuer_public_key(public_key), False)
        .add_extension(x509.BasicConstraints(ca=True, path_length=None), True)
        .add_extension(x509.SubjectKeyIdentifier.from_public_key(public_key), False)
    )

    return builder.sign(private_key, algorithm)


def _ca(
    name: str, private_key: CertificateIssuerPrivateKeyTypes, certificate: x509.Certificate
) -> CertificateAuthority:
    ca = CertificateAuthority(
        name=name, key_backend_alias="default", ocsp_key_backend_alias="default"
    )
    ca.key_backend.store_private_key(ca, private_key, certificate, DBStorePrivateKeyOptions())
    ca.update_certificate(certificate)
    ca.save()
    return ca


def generate_ec_private_key_fixture(
    elliptic_curve: ec.EllipticCurve,
) -> Callable[[], ec.EllipticCurvePrivateKey]:
    """Function to generate a fixture for an elliptic curve private key."""

    @pytest.fixture(scope="session")
    def func() -> ec.EllipticCurvePrivateKey:
        return ec.generate_private_key(elliptic_curve)

    return func


def generate_ec_certificate_fixture(
    curve: ec.EllipticCurve,
) -> Callable[["SubRequest"], x509.Certificate]:
    """Function to generate a fixture for an elliptic curve certificate."""

    @pytest.fixture(scope="session")
    def func(request: "SubRequest") -> x509.Certificate:
        private_key = request.getfixturevalue(f"ec_private_key_{curve.name}")
        return _sign(private_key, f"ec_{curve.name}", hashes.SHA256())

    return func


def generate_rsa_private_key_fixture(key_size: int) -> Callable[[], rsa.RSAPrivateKey]:
    """Function to generate a fixture for an RSA private key."""

    @pytest.fixture(scope="session")
    def func() -> rsa.RSAPrivateKey:
        return rsa.generate_private_key(public_exponent=65537, key_size=key_size)

    return func


def generate_rsa_certificate_fixture(
    key_size: int, algorithm: AllowedHashTypes
) -> Callable[[SubRequest], x509.Certificate]:
    """Function to generate a fixture for an RSA certificate."""

    @pytest.fixture(scope="session")
    def func(request: "SubRequest") -> x509.Certificate:
        private_key = request.getfixturevalue(f"rsa_private_key_{key_size}")
        return _sign(private_key, f"rsa_{key_size}", algorithm)

    return func


def generate_ca_fixture(
    name: str, private_key_name: str, certificate_name: str
) -> Callable[[SubRequest], CertificateAuthority]:
    """Function to generate a CA fixture."""

    @pytest.fixture
    def func(request: "SubRequest") -> CertificateAuthority:
        """Fixture for RSA-based certificate authorities."""
        request.getfixturevalue("db")
        private_key: ec.EllipticCurvePrivateKey = request.getfixturevalue(private_key_name)
        certificate: x509.Certificate = request.getfixturevalue(certificate_name)

        # Some sanity checks that private key/certificate match
        if isinstance(private_key, ec.EllipticCurvePrivateKey):
            ec_public_key = cast(ec.EllipticCurvePublicKey, certificate.public_key())
            assert isinstance(ec_public_key, ec.EllipticCurvePublicKey)
            assert private_key.curve == ec_public_key.curve
            assert private_key.key_size == ec_public_key.key_size
        elif isinstance(private_key, rsa.RSAPrivateKey):
            rsa_public_key = cast(rsa.RSAPublicKey, certificate.public_key())
            assert isinstance(rsa_public_key, rsa.RSAPublicKey)
            assert private_key.key_size == rsa_public_key.key_size

        return _ca(name, private_key, certificate)

    return func


@pytest.fixture(scope="session")
def dsa_private_key() -> dsa.DSAPrivateKey:
    """Session fixture for a DSA private key."""
    return dsa.generate_private_key(key_size=1024)


@pytest.fixture(scope="session")
def dsa_certificate(dsa_private_key: dsa.DSAPrivateKey) -> x509.Certificate:
    """Session fixture for a signed certificate of a DSA CA."""
    return _sign(dsa_private_key, "dsa", hashes.SHA256())


@pytest.fixture(scope="session")
def ed448_private_key() -> ed448.Ed448PrivateKey:
    """Session fixture for an Ed448 private key."""
    return ed448.Ed448PrivateKey.generate()


@pytest.fixture(scope="session")
def ed448_certificate(ed448_private_key: ed448.Ed448PrivateKey) -> x509.Certificate:
    """Session fixture for a signed certificate of an Ed448 CA."""
    return _sign(ed448_private_key, "ed448")


@pytest.fixture
def ed448_ca(
    request: "SubRequest",
    ed448_private_key: ed448.Ed448PrivateKey,
    ed448_certificate: x509.Certificate,
) -> CertificateAuthority:
    """Fixture for an Ed448 CA."""
    request.getfixturevalue("db")
    return _ca("ed448", ed448_private_key, ed448_certificate)


@pytest.fixture(scope="session")
def ed25519_private_key() -> ed25519.Ed25519PrivateKey:
    """Session fixture for an Ed25519 private key."""
    return ed25519.Ed25519PrivateKey.generate()


@pytest.fixture(scope="session")
def ed25519_certificate(ed25519_private_key: ed25519.Ed25519PrivateKey) -> x509.Certificate:
    """Session fixture for a signed certificate of an Ed25519 CA."""
    return _sign(ed25519_private_key, "ed25519")


@pytest.fixture
def ed25519_ca(
    request: "SubRequest",
    ed25519_private_key: ed25519.Ed25519PrivateKey,
    ed25519_certificate: x509.Certificate,
) -> CertificateAuthority:
    """Fixture for an Ed25519 CA."""
    request.getfixturevalue("db")
    return _ca("ed25519", ed25519_private_key, ed25519_certificate)


for _curve in ELLIPTIC_CURVES:
    globals()[f"ec_private_key_{_curve.name}"] = generate_ec_private_key_fixture(_curve)
    globals()[f"ec_certificate_{_curve.name}"] = generate_ec_certificate_fixture(_curve)
    globals()[f"ec_{_curve.name}_ca"] = generate_ca_fixture(
        f"ec_{_curve.name}", f"ec_private_key_{_curve.name}", f"ec_certificate_{_curve.name}"
    )
for _key_size in RSA_KEY_SIZES:
    globals()[f"rsa_private_key_{_key_size}"] = generate_rsa_private_key_fixture(_key_size)
    for _algorithm in RSA_ALGORITHMS:
        globals()[f"rsa_certificate_{_key_size}_{_algorithm.name}"] = (
            generate_rsa_certificate_fixture(_key_size, _algorithm)
        )
        globals()[f"rsa_{_key_size}_{_algorithm.name}_ca"] = generate_ca_fixture(
            f"rsa_{_key_size}_{_algorithm.name}",
            f"rsa_private_key_{_key_size}",
            f"rsa_certificate_{_key_size}_{_algorithm.name}",
        )


@pytest.fixture(params=[f"ec_{curve.name}_ca" for curve in ELLIPTIC_CURVES])
def ec_ca(request: "SubRequest") -> CertificateAuthority:
    """Fixture for all elliptic curve-based CAs."""
    ca = request.getfixturevalue(request.param)
    assert isinstance(ca, CertificateAuthority)
    return ca


@pytest.fixture(
    params=[
        f"rsa_{key_size}_{algorithm.name}_ca"
        for key_size in RSA_KEY_SIZES
        for algorithm in RSA_ALGORITHMS
    ]
    + [f"ec_{curve.name}_ca" for curve in ELLIPTIC_CURVES]
    + ["ed448_ca", "ed25519_ca"]
)
def ca(request: "SubRequest") -> CertificateAuthority:
    """Fixture for all RSA, elliptic curve, Ed448 and Ed25519 based CAs."""
    ca = request.getfixturevalue(request.param)
    assert isinstance(ca, CertificateAuthority)
    return ca


@pytest.fixture
def pre_created_client(ca: CertificateAuthority) -> CMCClient:
    """Create CMCClient with certificate for pre-created requests."""
    client = CMCClient()
    cert = x509.load_pem_x509_certificate(CMC_CLIENT_ONE_PEM)

    client.update_certificate(cert)
    client.save()
    return client


@pytest.fixture(scope="session")
def cmc_client_private_key() -> ec.EllipticCurvePrivateKey:
    """CMC client private key generated for this session."""
    return ec.generate_private_key(ec.SECP256R1())


@pytest.fixture(scope="session")
def cmc_client_public_key(
    cmc_client_private_key: ec.EllipticCurvePrivateKey,
) -> x509.Certificate:
    """CMC client public key generated for this session."""
    return _sign(cmc_client_private_key, "cmc-client.example.com", hashes.SHA256(), ca=False)


@pytest.fixture
def cmc_client(request: "SubRequest", cmc_client_public_key: x509.Certificate) -> CMCClient:
    """CMCClient instance with public key generated for this session."""
    request.getfixturevalue("db")
    client = CMCClient()
    client.update_certificate(cmc_client_public_key)
    client.save()
    return client



# Generated by Django 5.1.6 on 2025-02-21 18:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca_cmc', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='cmcclient',
            options={'verbose_name': 'CMC client', 'verbose_name_plural': 'CMC clients'},
        ),
        migrations.AddField(
            model_name='cmcclient',
            name='copy_extensions',
            field=models.BooleanField(default=False, help_text='Copy (almost) all extensions from the CSR if a request is signed by this certificate.'),
        ),
    ]



"""Test utility functions."""

from unittest.mock import PropertyMock, patch

import asn1crypto.cms
import freezegun
import pytest
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, rsa
from python_cmc import cmc

from django_ca_cmc.utils import get_signed_digest_algorithm
from tests.utils import create_cmc_request, create_tagged_csr_message, load_file


def test_get_signed_digest_algorithm_with_rsa_with_unsupported_hash_algorithm(
    rsa_certificate_2048_sha256: x509.Certificate,
) -> None:
    """Test unsupported hash algorithm."""
    mock = PropertyMock(return_value=hashes.SHA3_256())
    error = r"^sha3-256: Signature hash algorithm not supported\.$"
    with patch("cryptography.x509.Certificate.signature_hash_algorithm", new_callable=mock):
        with pytest.raises(ValueError, match=error):
            get_signed_digest_algorithm(rsa_certificate_2048_sha256)


def test_get_signed_digest_algorithm_with_ec_with_unsupported_curve(
    ec_certificate_sect571k1: x509.Certificate,
) -> None:
    """Test unsupported curve."""
    with pytest.raises(ValueError, match=r"^sect571k1: Elliptic curve not supported\.$"):
        get_signed_digest_algorithm(ec_certificate_sect571k1)


def test_get_signed_digest_algorithm_with_unsupported_object() -> None:
    """Test unsupported object."""
    error = r"None: Must be of type cryptography\.x509\.Certificate\.$"
    with pytest.raises(TypeError, match=error):
        get_signed_digest_algorithm(None)  # type: ignore[arg-type]  # What we're testing.


def test_get_signed_digest_algorithm_with_unsopported_public_key_type(
    dsa_certificate: x509.Certificate,
) -> None:
    """Test unsupported public key type (DSA)."""
    with pytest.raises(ValueError, match=r"Public key type not supported\.$"):
        get_signed_digest_algorithm(dsa_certificate)


def test_create_tagged_csr_message(rsa_private_key_2048: rsa.RSAPrivateKey) -> None:
    """Test create_tagged_csr_message()."""
    expected_msg = asn1crypto.cms.ContentInfo.load(load_file("cmc_with_csr"))
    raw_cmc_request = expected_msg["content"]["encap_content_info"]["content"].parsed.dump()
    expected = cmc.PKIData.load(raw_cmc_request)

    # Get CSR from expected request
    csr = x509.load_der_x509_csr(expected["reqSequence"][0].chosen["certificationRequest"].dump())

    actual = cmc.PKIData.load(create_tagged_csr_message(csr))
    assert expected.native == actual.native


def test_create_cmc_request(
    cmc_client_private_key: ec.EllipticCurvePrivateKey,
    cmc_client_public_key: x509.Certificate,
) -> None:
    """Test the create_cmc_request function."""
    expected_msg = asn1crypto.cms.ContentInfo.load(load_file("cmc_with_csr"))
    data = expected_msg["content"]["encap_content_info"]["content"].native

    with freezegun.freeze_time("2023-1-30 22:18:43"):
        actual_msg = asn1crypto.cms.ContentInfo.load(
            create_cmc_request(cmc_client_private_key, cmc_client_public_key, data)
        )

    # Just ensure that data is generally valid
    _ = expected_msg.native
    _ = actual_msg.native

    # Verify content type
    assert actual_msg["content_type"] == expected_msg["content_type"]

    # Retrieve content
    expected = expected_msg["content"]
    actual = actual_msg["content"]
    assert isinstance(expected, asn1crypto.cms.SignedData)
    assert isinstance(actual, asn1crypto.cms.SignedData)

    # Assert basic properties of ContentInfo
    assert actual["version"] == expected["version"]
    assert actual["digest_algorithms"].native == expected["digest_algorithms"].native
    assert isinstance(actual["certificates"], type(expected["certificates"]))
    assert isinstance(actual["certificates"][0], type(expected["certificates"][0]))
    assert isinstance(actual["signer_infos"], type(expected["signer_infos"]))
    assert isinstance(actual["signer_infos"][0], type(expected["signer_infos"][0]))

    # Check the encapsulated content info
    actual_encap_content_info = actual["encap_content_info"]
    expected_encap_content_info = expected["encap_content_info"]
    assert isinstance(actual_encap_content_info, asn1crypto.cms.EncapsulatedContentInfo)
    assert isinstance(expected_encap_content_info, asn1crypto.cms.EncapsulatedContentInfo)
    assert actual_encap_content_info["content_type"] == expected_encap_content_info["content_type"]
    print(expected_encap_content_info["content"].native)
    assert (
        actual_encap_content_info["content"].native == expected_encap_content_info["content"].native
    )

    # Check signer_info
    assert len(actual["signer_infos"]) == len(expected["signer_infos"]) == 1
    actual_signer_info = actual["signer_infos"][0]
    expected_signer_info = expected["signer_infos"][0]
    assert isinstance(actual_signer_info, asn1crypto.cms.SignerInfo)
    assert isinstance(expected_signer_info, asn1crypto.cms.SignerInfo)

    assert actual_signer_info["version"].native == expected_signer_info["version"].native
    assert (
        actual_signer_info["digest_algorithm"].native
        == expected_signer_info["digest_algorithm"].native
    )
    assert (
        actual_signer_info["signature_algorithm"].native
        == expected_signer_info["signature_algorithm"].native
    )
    assert isinstance(actual_signer_info["signature"].native, bytes)
    assert isinstance(actual_signer_info["sid"], asn1crypto.cms.SignerIdentifier)

    # verify signer infos
    assert actual_signer_info["signed_attrs"].native == expected_signer_info["signed_attrs"].native



"""Constants used in various places throughout the code."""

from cryptography.hazmat.primitives.asymmetric import ec, ed448, ed25519, rsa

#: Tuple of supported public key types
SUPPORTED_PUBLIC_KEY_TYPES = (
    rsa.RSAPublicKey | ec.EllipticCurvePublicKey | ed448.Ed448PublicKey | ed25519.Ed25519PublicKey
)



"""URL configuration for the django-ca-cmc app."""

from django.urls import URLPattern, URLResolver, path

from django_ca_cmc import views

app_name = "django_ca_cmc"

urlpatterns: list[URLResolver | URLPattern] = [
    path("<serial:serial>/", views.CMCView.as_view(), name="cmc"),
]



"""Admin for django-ca-cmc."""

from typing import TYPE_CHECKING, ClassVar

# Register your models here.
from django.contrib import admin

from django_ca_cmc.models import CMCClient

if TYPE_CHECKING:
    CMCClientAdminBase = admin.ModelAdmin[CMCClient]
else:
    CMCClientAdminBase = admin.ModelAdmin


@admin.register(CMCClient)
class CMCClientAdmin(CMCClientAdminBase):
    """Model admin class for CMC client."""

    list_display = ("serial", "not_before", "not_after")
    readonly_fields = ("not_after", "not_before", "serial")

    class Media:
        css: ClassVar[dict[str, tuple[str, ...]]] = {
            "all": ("django_ca/admin/css/base.css",),
        }



"""Models for django-ca-cmc."""

from cryptography import x509
from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _
from django_ca.modelfields import CertificateField, LazyCertificate
from django_ca.utils import int_to_hex


class CMCClient(models.Model):
    """CMC client certificate."""

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    # Key client data
    certificate = CertificateField(verbose_name=_("Client certificate"))

    # Properties of the certificate for faster data access
    not_before = models.DateTimeField(blank=False)
    not_after = models.DateTimeField(null=False, blank=False)
    serial = models.CharField(max_length=64, unique=True)

    comment = models.TextField(default="")
    copy_extensions = models.BooleanField(
        default=False,
        help_text=_(
            "Copy (almost) all extensions from the CSR if a request is signed by this certificate."
        ),
    )

    class Meta:
        verbose_name = _("CMC client")
        verbose_name_plural = _("CMC clients")

    def __str__(self) -> str:
        return self.serial

    def update_certificate(self, value: x509.Certificate) -> None:
        """
        Update instance with data from a :py:class:`cg:cryptography.x509.Certificate`.

        This function will also populate the ``serial, `not_after` and `not_before` fields.
        """
        self.certificate = LazyCertificate(value)
        self.not_after = value.not_valid_after_utc
        self.not_before = value.not_valid_before_utc

        if settings.USE_TZ is False:
            self.not_after = self.not_after.replace(tzinfo=None)
            self.not_before = self.not_before.replace(tzinfo=None)

        self.serial = int_to_hex(value.serial_number)  # upper-cased by int_to_hex()



"""CMC management command."""

import json
from typing import Any, Literal

from django.core.management import CommandParser
from django.utils import timezone
from django_ca.management.base import BaseCommand
from tabulate import tabulate

from django_ca_cmc.models import CMCClient


class Command(BaseCommand):
    """Command class."""

    help = "List CMC client certificates."

    def add_arguments(self, parser: CommandParser) -> None:
        """Add arguments."""
        parser.add_argument(
            "-f",
            "--format",
            choices=("json", "table"),
            default="table",
            help="Select output format (default: %(default)s).",
        )
        parser.add_argument("-e", "--expired", help="Include clients with expired certificates.")

    def handle(self, format: Literal["json", "table"], expired: bool, **options: Any) -> None:
        """Handle method."""
        clients = CMCClient.objects.all()
        if not expired:
            now = timezone.now()
            clients = clients.filter(not_after__gt=now)

        if format == "json":
            data = [
                {
                    "serial": c.serial,
                    "not_before": c.not_before.isoformat(),
                    "not_after": c.not_after.isoformat(),
                    "comment": c.comment,
                    "pem": c.certificate.pem,
                }
                for c in clients
            ]
            print(json.dumps(data, sort_keys=True, indent=4))
        else:
            headers = ["serial", "not before", "not after", "comment"]
            table = [
                [c.serial, c.not_before.isoformat(), c.not_after.isoformat(), c.comment]
                for c in clients
            ]
            print(tabulate(table, headers, tablefmt="grid"))



"""App configuration for django-ca-cmc."""

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DjangoCaCmcConfig(AppConfig):
    """Main Django app config."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "django_ca_cmc"
    verbose_name = _("Certificate Authority: CMC")



# Generated by Django 5.1.5 on 2025-02-07 21:59

import django_ca.modelfields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CMCClient',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('certificate', django_ca.modelfields.CertificateField(verbose_name='Client certificate')),
                ('not_before', models.DateTimeField()),
                ('not_after', models.DateTimeField()),
                ('serial', models.CharField(max_length=64, unique=True)),
                ('comment', models.TextField(default='')),
            ],
        ),
    ]



"""Settings for django-ca-cmc."""

from typing import Annotated, Any

from cryptography import x509
from cryptography.x509.oid import ExtensionOID
from django_ca.conf import SettingsProxyBase
from django_ca.pydantic.type_aliases import Serial
from pydantic import BaseModel, BeforeValidator, ConfigDict


def oid_validator(value: Any) -> Any:
    """Convert OID strings to x509.ObjectIdentifier."""
    print(value, type(value))
    if isinstance(value, str):
        return x509.ObjectIdentifier(value)
    return value


OID = Annotated[x509.ObjectIdentifier, BeforeValidator(oid_validator)]


class CMCModelSettings(BaseModel):
    """CMC settings."""

    model_config = ConfigDict(arbitrary_types_allowed=True, from_attributes=True, frozen=True)

    CA_CMC_COPY_CSR_EXTENSIONS_BLACKLIST: tuple[OID, ...] = (
        ExtensionOID.AUTHORITY_INFORMATION_ACCESS,  # contains OCSP URLs
        ExtensionOID.AUTHORITY_KEY_IDENTIFIER,
        ExtensionOID.BASIC_CONSTRAINTS,
        ExtensionOID.CRL_DISTRIBUTION_POINTS,
        ExtensionOID.SUBJECT_KEY_IDENTIFIER,
    )
    CA_CMC_COPY_UNRECOGNIZED_CSR_EXTENSIONS: bool = False
    CA_CMC_DEFAULT_SERIAL: Serial | None = None
    CA_CMC_DEFAULT_RESPONDER_SERIAL: Serial | None = None


class CMCSettingsProxy(SettingsProxyBase[CMCModelSettings]):
    """Proxy class to access model settings."""

    settings_model = CMCModelSettings
    __settings: CMCModelSettings


cmc_settings = CMCSettingsProxy()



"""Views for django-ca-cmc."""

import logging

import asn1crypto.cms
import asn1crypto.x509
from cryptography import x509
from cryptography.exceptions import InvalidSignature
from django.core.exceptions import BadRequest, ImproperlyConfigured, PermissionDenied
from django.http import HttpRequest, HttpResponse, HttpResponseBadRequest, HttpResponseForbidden
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from django_ca.models import CertificateAuthority
from python_cmc import cmc

from django_ca_cmc.cmc import (
    check_request_signature,
    create_cert_from_csr,
    create_cmc_response,
    create_csr_from_crmf,
)
from django_ca_cmc.conf import cmc_settings

log = logging.getLogger(__name__)
CONTENT_TYPE = "application/pkcs7-mime"


@method_decorator(csrf_exempt, name="dispatch")
class CMCView(View):
    """View handling CMC requests."""

    serial: str | None = None
    responder_serial: str | None = None

    def handle_request(
        self, ca: CertificateAuthority, responder_authority: CertificateAuthority, data: bytes
    ) -> bytes:
        """Handle and extract CMS CMC request."""
        created_certs: dict[int, asn1crypto.x509.Certificate] = {}

        content_info = asn1crypto.cms.ContentInfo.load(data)
        _ = content_info.native  # Ensure valid data
        content = content_info["content"]

        # Ensure valid signature for the request
        if len(content["signer_infos"]) == 0 or len(content["certificates"]) == 0:
            raise PermissionDenied("Invalid signature or certificate for the signature")
        client = check_request_signature(content["certificates"], content["signer_infos"])

        raw_cmc_request = content["encap_content_info"]["content"].parsed.dump()
        cmc_req = cmc.PKIData.load(raw_cmc_request)
        _ = cmc_req.native  # Ensure valid data

        cmc_requests = [req.chosen for req in cmc_req["reqSequence"]]

        # Make sure that all sub-requests are supported
        if unsupported := [
            _
            for _ in cmc_requests
            if not isinstance(_, cmc.CertReqMsg | cmc.TaggedCertificationRequest)
        ]:
            log.error("CMC requests of unknown type: %s", unsupported)
            raise BadRequest("CMC requests of unknown type.")

        try:
            for value in cmc_requests:
                if isinstance(value, cmc.CertReqMsg):  # CRMF
                    req_id = int(value["certReq"]["certReqId"].native)
                    csr = create_csr_from_crmf(value["certReq"]["certTemplate"])
                elif isinstance(value, cmc.TaggedCertificationRequest):  # CSR
                    req_id = int(value["bodyPartID"].native)
                    csr = x509.load_der_x509_csr(value["certificationRequest"].dump())
                else:  # pragma: no cover  # must not happen, types asserted outside of loop
                    raise ValueError("Unsupported message type")

                created_certs[req_id] = create_cert_from_csr(ca, client, csr)

            ret = create_cmc_response(
                ca, responder_authority, cmc_req["controlSequence"], created_certs, failed=False
            )
        except (ValueError, TypeError):
            ret = create_cmc_response(
                ca, responder_authority, cmc_req["controlSequence"], created_certs, failed=True
            )

        return ret

    def get(self, request: HttpRequest, serial: str | None = None) -> HttpResponse:
        return HttpResponse("CMC endpoint here!")

    def get_certificate_authority(self, serial: str | None) -> CertificateAuthority:
        if serial is None:
            serial = self.serial
        if serial is None:
            serial = cmc_settings.CA_CMC_DEFAULT_SERIAL
        if serial is None:
            # If it's still None, we cannot determine the serial.
            raise ImproperlyConfigured("No serial configured for this view.")

        return CertificateAuthority.objects.usable().get(serial=serial)

    def get_responder_authority(self, ca: CertificateAuthority) -> CertificateAuthority:
        serial = self.responder_serial
        if serial is None:
            serial = cmc_settings.CA_CMC_DEFAULT_RESPONDER_SERIAL
        if serial is None:
            serial = ca.serial

        if serial == ca.serial:
            return ca
        else:
            return CertificateAuthority.objects.usable().get(serial=serial)

    def post(self, request: HttpRequest, serial: str | None = None) -> HttpResponse:
        content_type = request.headers.get("Content-type")
        if content_type is None or content_type != CONTENT_TYPE:
            return HttpResponseBadRequest("invalid content type", content_type=CONTENT_TYPE)

        certificate_authority = self.get_certificate_authority(serial)
        responder_authority = self.get_responder_authority(certificate_authority)

        try:
            data_content = self.handle_request(
                certificate_authority, responder_authority, request.body
            )
        except InvalidSignature:
            return HttpResponseForbidden("invalid signature", content_type=CONTENT_TYPE)
        except Exception as ex:
            log.exception("Internal error: %s", ex)
            return HttpResponseBadRequest("internal error", content_type=CONTENT_TYPE)

        return HttpResponse(data_content, content_type=CONTENT_TYPE)



"""Django settings for unit tests."""

from cryptography.x509.oid import ExtensionOID

SECRET_KEY = "dummy"
ROOT_URLCONF = "tests.urls"
INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django_ca",
    "django_ca_cmc",
]
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    },
}

CA_MIN_KEY_SIZE = 1024
CA_KEY_BACKENDS = {
    "default": {
        "BACKEND": "django_ca.key_backends.db.DBBackend",
    },
}

CA_CMC_COPY_CSR_EXTENSIONS = (ExtensionOID.CERTIFICATE_POLICIES.dotted_string,)



"""Collect functions related to hashing."""

import hashlib

import asn1crypto.algos


def digest(data: bytes, algorithm: asn1crypto.algos.DigestAlgorithm) -> bytes:
    """
    Hash arbitrary data using the given digest algorithm.

    This function supports the full range of SHA2 and SHA3 algorithms.
    """
    algorithm = algorithm["algorithm"].native

    if algorithm in ("md5", "sha1"):
        raise ValueError(f"{algorithm}: Cannot hash data using this algorithm")

    hasher = hashlib.new(algorithm, data)
    try:
        return hasher.digest()
    except TypeError as ex:
        raise ValueError(f"{algorithm}: Unable to use algorithm for hashing") from ex



"""CMC-related functions."""

import logging
import secrets
from collections.abc import Sequence
from datetime import UTC, datetime

import asn1crypto.algos
import asn1crypto.keys
import asn1crypto.pem
import asn1crypto.x509
from asn1crypto import cms
from cryptography import x509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, ed448, ed25519, padding, rsa
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.asymmetric.types import CertificateIssuerPublicKeyTypes
from django.conf import settings
from django_ca.models import Certificate, CertificateAuthority, X509CertMixin
from python_cmc import cmc

from django_ca_cmc.conf import cmc_settings
from django_ca_cmc.constants import SUPPORTED_PUBLIC_KEY_TYPES
from django_ca_cmc.hash import digest
from django_ca_cmc.models import CMCClient
from django_ca_cmc.utils import get_signed_digest_algorithm

log = logging.getLogger(__name__)


def _public_key_verify_ecdsa_signature(
    pub_key: ec.EllipticCurvePublicKey, signature: bytes, signed_data: bytes
) -> None:
    # NOTE: We removed convert_rs_ec_signature conversion here. The signature comes from the
    #   sender and hopefully does not really still require conversion. If signature verification
    #   fails, we might have to re-add this.
    if pub_key.curve.name == "secp256r1":
        pub_key.verify(signature, signed_data, ec.ECDSA(hashes.SHA256()))
    elif pub_key.curve.name == "secp384r1":
        pub_key.verify(signature, signed_data, ec.ECDSA(hashes.SHA384()))
    elif pub_key.curve.name == "secp521r1":
        pub_key.verify(signature, signed_data, ec.ECDSA(hashes.SHA512()))
    else:
        raise ValueError("Unsupported EC curve")


def verify_signature(
    public_key: CertificateIssuerPublicKeyTypes, signature: bytes, signed_data: bytes
) -> None:
    """
    Verify signature with a public key.

    raises cryptography.exceptions.InvalidSignature
    if invalid signature or ValueError if the public key is not supported.

    Potentially fails if the signature is made using nonstandard hashing of the data.
    """
    if isinstance(public_key, rsa.RSAPublicKey):
        try:
            public_key.verify(signature, signed_data, padding.PKCS1v15(), hashes.SHA256())
        except InvalidSignature:
            public_key.verify(signature, signed_data, padding.PKCS1v15(), hashes.SHA512())

    elif isinstance(public_key, ec.EllipticCurvePublicKey):
        _public_key_verify_ecdsa_signature(public_key, signature, signed_data)

    elif isinstance(public_key, (ed25519.Ed25519PublicKey | ed448.Ed448PublicKey)):
        public_key.verify(signature, signed_data)

    else:
        raise ValueError("Non supported public key in certificate")


def check_request_signature(
    request_signers: cms.CertificateSet, signer_infos: cms.SignerInfos
) -> CMCClient:
    """Check a CMC request signature."""
    now = datetime.now(tz=UTC)
    if settings.USE_TZ is False:
        now = now.replace(tzinfo=None)

    clients = CMCClient.objects.filter(not_before__lt=now, not_after__gt=now)

    for request_signer in request_signers:
        for client in clients:
            cert = asn1crypto.x509.Certificate.load(client.certificate.der)
            if request_signer.chosen.native == cert.native:
                for signer_info in signer_infos:
                    signer_cert = x509.load_der_x509_certificate(request_signer.chosen.dump())
                    public_key = signer_cert.public_key()
                    if not isinstance(public_key, SUPPORTED_PUBLIC_KEY_TYPES):
                        raise ValueError(
                            f"{public_key}: Client passed an unsupported public signer key."
                        )

                    signature = signer_info["signature"].contents
                    signed_data = signer_info["signed_attrs"].retag(17).dump()
                    try:
                        verify_signature(public_key, signature, signed_data)
                        return client
                    except (InvalidSignature, ValueError, TypeError):
                        pass

    raise InvalidSignature("Wrong or missing CMS signer")


# FIXME This need to be improved to fully support crmf, not just basics
def create_csr_from_crmf(
    certificate_request_message: cmc.CertReqMsg,
) -> x509.CertificateSigningRequest:
    """Manually handle the CRMF request into a CSR."""
    attrs = asn1crypto.csr.CRIAttributes()

    cert_req_info = asn1crypto.csr.CertificationRequestInfo()
    cert_req_info["version"] = 0
    cert_req_info["subject"] = certificate_request_message["subject"]
    cert_req_info["subject_pk_info"] = certificate_request_message["publicKey"]

    set_of_exts = asn1crypto.csr.SetOfExtensions()

    set_of_exts.append(certificate_request_message["extensions"])
    cri_attr = asn1crypto.csr.CRIAttribute(
        {"type": asn1crypto.csr.CSRAttributeType("1.2.840.113549.1.9.14"), "values": set_of_exts}
    )
    attrs.append(cri_attr)

    cert_req_info["attributes"] = attrs

    asn1crypto_certification_request = asn1crypto.csr.CertificationRequest()
    asn1crypto_certification_request["certification_request_info"] = cert_req_info
    asn1crypto_certification_request["signature_algorithm"] = (
        asn1crypto.algos.SignedDigestAlgorithm(
            {"algorithm": asn1crypto.algos.SignedDigestAlgorithmId("1.2.840.10045.4.3.2")}
        )
    )
    asn1crypto_certification_request["signature"] = b"dummy_sig"

    # Convert CSR to cryptography and return.
    return x509.load_der_x509_csr(asn1crypto_certification_request.dump())


def create_cert_from_csr(
    ca: CertificateAuthority, client: CMCClient, csr: x509.CertificateSigningRequest
) -> Certificate:
    """Create cert from a csr."""
    key_backend_options = ca.key_backend.get_use_private_key_options(ca, {})

    extensions = []
    if client.copy_extensions is True:
        extensions = [
            ext
            for ext in csr.extensions
            if ext.oid not in cmc_settings.CA_CMC_COPY_CSR_EXTENSIONS_BLACKLIST
        ]

    return Certificate.objects.create_cert(
        ca,
        key_backend_options,
        csr,
        subject=csr.subject,
        extensions=extensions,
        allow_unrecognized_extensions=cmc_settings.CA_CMC_COPY_UNRECOGNIZED_CSR_EXTENSIONS,
    )


def cmc_revoke(revoke_data: bytes) -> None:
    """Revoke a certificate based on the CMC RevokeRequest."""
    # set_of_revoke_request = cmc.SetOfRevokeRequest.load(revoke_data)
    # revoked_certs = 0
    #
    # for revoke_request in set_of_revoke_request:
    #     # Try certs
    #     db_certificate_objs = await db_load_data_class(
    #         Certificate, CertificateInput(
    #           serial_number=str(revoke_request["serial_number"].native)
    #         )
    #     )
    #     for obj in db_certificate_objs:
    #         if isinstance(obj, Certificate):
    #             if (
    #                 pem_cert_to_name_dict(await obj.issuer_pem())
    #                 == revoke_request["issuerName"].native
    #             ):
    #                 await obj.revoke(
    #                     1, int(revoke_request["reason"])
    #                 )  # Change to cmc request signer
    #                 revoked_certs += 1
    #                 print("Revoked cert due to CMC request")
    #
    #     # Try Ca's
    #     db_ca_objs = await db_load_data_class(
    #         Ca, CaInput(serial_number=str(revoke_request["serial_number"].native))
    #     )
    #     for obj in db_ca_objs:
    #         if isinstance(obj, Ca):
    #             if (
    #                 pem_cert_to_name_dict(await obj.issuer_pem())
    #                 == revoke_request["issuerName"].native
    #             ):
    #                 await obj.revoke(
    #                     1, int(revoke_request["reason"])
    #                 )  # Change to cmc request signer
    #                 revoked_certs += 1
    #                 print("Revoked cert due to CMC request")
    #
    # if revoked_certs == 0:
    #     print("Could not find the certificate to revoke from CMC RevokeRequest")
    #     raise ValueError


def _create_cmc_response_status_packet(
    created_certs: dict[int, asn1crypto.x509.Certificate], failed: bool
) -> cmc.TaggedAttribute:
    body_part_references = cmc.BodyPartReferences()

    for req_id in created_certs:
        body_part_references.append(cmc.BodyPartReference({"bodyPartID": req_id}))

    status_v2 = cmc.CMCStatusInfoV2()
    if len(body_part_references) == 0:
        status_v2["bodyList"] = cmc.BodyPartReferences([])
    else:
        status_v2["bodyList"] = body_part_references

    if failed:
        status_v2["cMCStatus"] = cmc.CMCStatus(2)
        status_v2["statusString"] = "Failed processing CMC request"
        status_v2["otherInfo"] = cmc.OtherStatusInfo({"failInfo": cmc.CMCFailInfo(11)})
    else:
        status_v2["cMCStatus"] = cmc.CMCStatus(0)
        status_v2["statusString"] = "OK"

    status_v2_attr_values = cmc.SetOfCMCStatusInfoV2()
    status_v2_attr_values.append(status_v2)
    status_v2_attr = cmc.TaggedAttribute()
    status_v2_attr["bodyPartID"] = secrets.randbelow(4294967293)
    status_v2_attr["attrType"] = cmc.TaggedAttributeType("1.3.6.1.5.5.7.7.25")
    status_v2_attr["attrValues"] = status_v2_attr_values
    return status_v2_attr


def create_cmc_response_packet(
    controls: cmc.Controls, created_certs: dict[int, asn1crypto.x509.Certificate], failed: bool
) -> cmc.PKIResponse:
    """
    Create a CMC response package.

    Revoke cert(s) if the request had a RevokeRequest(s).
    """
    response_controls = cmc.Controls()
    nonce: bytes | None = None
    reg_info: bytes | None = None

    for control_value in controls:
        if control_value["attrType"].native == "id-cmc-senderNonce":
            nonce = control_value["attrValues"].dump()

        if control_value["attrType"].native == "id-cmc-regInfo":
            reg_info = control_value["attrValues"].dump()

    # If a revoke request
    if not failed:
        for control_value in controls:
            if control_value["attrType"].native == "id-cmc-revokeRequest":
                revoke_request = control_value["attrValues"].dump()
                cmc_revoke(revoke_request)

    if nonce is not None:
        nonce_attr = cmc.TaggedAttribute()
        nonce_attr["bodyPartID"] = secrets.randbelow(4294967293)
        nonce_attr["attrType"] = cmc.TaggedAttributeType("1.3.6.1.5.5.7.7.7")
        nonce_attr["attrValues"] = asn1crypto.cms.SetOfOctetString.load(nonce)
        response_controls.append(nonce_attr)

    if reg_info is not None:
        reg_info_attr = cmc.TaggedAttribute()
        reg_info_attr["bodyPartID"] = secrets.randbelow(4294967293)
        reg_info_attr["attrType"] = cmc.TaggedAttributeType("1.3.6.1.5.5.7.7.19")
        reg_info_attr["attrValues"] = asn1crypto.cms.SetOfOctetString.load(reg_info)
        response_controls.append(reg_info_attr)

    status_v2_attr = _create_cmc_response_status_packet(created_certs, failed)
    response_controls.append(status_v2_attr)

    pki_response = cmc.PKIResponse()
    pki_response["controlSequence"] = response_controls
    pki_response["cmsSequence"] = cmc.TaggedContentInfos([])
    pki_response["otherMsgSequence"] = cmc.OtherMsgs([])
    return pki_response


def pem_cert_to_key_hash(certificate: x509.Certificate) -> bytes:
    """Get digest of the SubjectKeyIdentifier extension of the given certificate."""
    try:
        ext: x509.Extension[x509.SubjectKeyIdentifier] = (
            certificate.extensions.get_extension_for_class(x509.SubjectKeyIdentifier)
        )
    except x509.ExtensionNotFound as ex:
        raise ValueError("No SubjectKeyIdentifier extension in certificate.") from ex

    return ext.value.digest


def create_cmc_response(
    ca: CertificateAuthority,
    responder_authority: CertificateAuthority,
    controls: cmc.Controls,
    created_certs: dict[int, Certificate],
    failed: bool,
) -> bytes:
    """Create a CMS response containing a CMC package."""
    # https://github.com/SUNET/pkcs11_ca/blob/main/src/pkcs11_ca_service/cmc.py#L207
    # Add CA bundle and created certificates to the chain.
    model_chain: Sequence[X509CertMixin] = ca.bundle + list(created_certs.values())

    # Get digest and signature algorithms
    # NOTE: pkcs11_ca_service *always* uses sha256 for digest algorithm.
    digest_algorithm_name = getattr(settings, "CA_CMC_DIGEST_ALGORITHM", "sha256")
    digest_algorithm = asn1crypto.algos.DigestAlgorithm(
        {"algorithm": asn1crypto.algos.DigestAlgorithmId(digest_algorithm_name)}
    )

    # NOTE: signature_algorithm is set twice in the response and one should *probably* be based on
    # the responder authority. However, pkcs11_ca_service also always uses the same value. Further
    # refinement here would require testing with actual clients.
    signature_algorithm = get_signed_digest_algorithm(ca.pub.loaded)

    # Convert chain to asn1crypto options
    chain: list[asn1crypto.x509.Certificate] = [
        asn1crypto.x509.Certificate.load(c.pub.der) for c in model_chain
    ]

    packet = create_cmc_response_packet(controls, created_certs, failed)

    eci = asn1crypto.cms.EncapsulatedContentInfo()
    eci["content_type"] = asn1crypto.cms.ContentType("1.3.6.1.5.5.7.12.3")
    packet_data = asn1crypto.core.ParsableOctetString()
    packet_data.set(packet.dump())
    eci["content"] = packet_data

    signed_data = asn1crypto.cms.SignedData()
    signed_data["version"] = 2
    signed_data["digest_algorithms"] = asn1crypto.cms.DigestAlgorithms({digest_algorithm})
    signed_data["encap_content_info"] = eci

    signer_info = asn1crypto.cms.SignerInfo()
    signer_info["version"] = 1
    signer_info["sid"] = asn1crypto.cms.SignerIdentifier(
        {"subject_key_identifier": pem_cert_to_key_hash(ca.pub.loaded)}
    )

    cms_attributes = asn1crypto.cms.CMSAttributes()
    cms_attributes.append(
        asn1crypto.cms.CMSAttribute(
            {
                "type": asn1crypto.cms.CMSAttributeType("content_type"),
                "values": asn1crypto.cms.SetOfContentType(
                    [asn1crypto.cms.ContentType("1.3.6.1.5.5.7.12.3")]
                ),
            }
        )
    )

    # Calculate message digest
    message_digest = digest(signed_data["encap_content_info"]["content"].contents, digest_algorithm)

    cms_attributes.append(
        asn1crypto.cms.CMSAttribute(
            {
                "type": asn1crypto.cms.CMSAttributeType("message_digest"),
                "values": asn1crypto.cms.SetOfOctetString([message_digest]),
            }
        )
    )

    cms_attributes.append(
        asn1crypto.cms.CMSAttribute(
            {
                "type": asn1crypto.cms.CMSAttributeType("signing_time"),
                "values": asn1crypto.cms.SetOfTime([asn1crypto.core.UTCTime(datetime.now(UTC))]),
            }
        )
    )

    cms_attributes.append(
        asn1crypto.cms.CMSAttribute(
            {
                "type": asn1crypto.cms.CMSAttributeType("cms_algorithm_protection"),
                "values": asn1crypto.cms.SetOfCMSAlgorithmProtection(
                    [
                        asn1crypto.cms.CMSAlgorithmProtection(
                            {
                                "digest_algorithm": digest_algorithm,
                                "signature_algorithm": signature_algorithm,
                            }
                        )
                    ]
                ),
            }
        )
    )

    signer_info["signed_attrs"] = cms_attributes

    signer_info["digest_algorithm"] = digest_algorithm
    signer_info["signature_algorithm"] = signature_algorithm

    # Sign the data
    raw_data = signer_info["signed_attrs"].retag(17).dump()

    response_padding = None
    if responder_authority.key_type == "RSA":
        response_padding = PKCS1v15()
    raw_signature = responder_authority.sign_data(raw_data, padding=response_padding)

    signer_info["signature"] = raw_signature

    signed_data["signer_infos"] = asn1crypto.cms.SignerInfos({signer_info})
    signed_data["certificates"] = asn1crypto.cms.CertificateSet(chain)

    cmc_resp = asn1crypto.cms.ContentInfo()
    cmc_resp["content_type"] = asn1crypto.cms.ContentType("signed_data")
    cmc_resp["content"] = signed_data

    ret: bytes = cmc_resp.dump()
    return ret



"""TEst the cmc_add_client command."""

from collections.abc import Iterator
from contextlib import contextmanager
from datetime import UTC, datetime
from importlib import resources
from io import StringIO
from unittest import mock

import pytest
from cryptography import x509
from cryptography.hazmat.primitives.serialization import Encoding
from django.core.management import CommandError, call_command

from django_ca_cmc.models import CMCClient
from tests import files
from tests.conftest import CMC_CLIENT_ONE_PEM

cmc_client_1_path = resources.files(files) / "cmc_client_1.pem"
serial = "617C352A"


@contextmanager
def stdin_bytes(value: bytes) -> Iterator[None]:
    """Mock stdin bytes."""

    # mock https://docs.python.org/3/library/io.html#io.BufferedReader.read
    def _read_mock(size=None):  # type: ignore # pylint: disable=unused-argument
        return value

    with mock.patch("sys.stdin.buffer.read", side_effect=_read_mock):
        yield


@pytest.mark.django_db
def test_with_path_with_pem() -> None:
    """Test passing a path to a PEM."""
    stdout = StringIO()
    call_command("cmc_add_client", cmc_client_1_path, stdout=stdout)
    assert stdout.getvalue() == f"Client with serial {serial} added successfully.\n"

    client = CMCClient.objects.get()
    assert client.certificate.pem.strip() == CMC_CLIENT_ONE_PEM.decode()
    assert client.serial == serial
    assert client.not_after == datetime(2026, 10, 29, 17, 53, 46, tzinfo=UTC)
    assert client.not_before == datetime(2021, 10, 29, 17, 53, 46, tzinfo=UTC)
    assert client.comment == ""
    assert client.copy_extensions is False


@pytest.mark.django_db
def test_with_parameters() -> None:
    """Test passing a path to a PEM."""
    stdout = StringIO()
    call_command(
        "cmc_add_client", cmc_client_1_path, comment="foo", copy_extensions=True, stdout=stdout
    )
    assert stdout.getvalue() == f"Client with serial {serial} added successfully.\n"

    client = CMCClient.objects.get()
    assert client.certificate.pem.strip() == CMC_CLIENT_ONE_PEM.decode()
    assert client.serial == serial
    assert client.not_after == datetime(2026, 10, 29, 17, 53, 46, tzinfo=UTC)
    assert client.not_before == datetime(2021, 10, 29, 17, 53, 46, tzinfo=UTC)
    assert client.comment == "foo"
    assert client.copy_extensions is True


@pytest.mark.django_db
def test_with_stdin_with_pem() -> None:
    """Test passing a path to a PEM."""
    stdout = StringIO()
    with stdin_bytes(CMC_CLIENT_ONE_PEM):
        call_command("cmc_add_client", "-", stdout=stdout)
    assert stdout.getvalue() == f"Client with serial {serial} added successfully.\n"

    client = CMCClient.objects.get()
    assert client.certificate.pem.strip() == CMC_CLIENT_ONE_PEM.decode()
    assert client.serial == serial
    assert client.not_after == datetime(2026, 10, 29, 17, 53, 46, tzinfo=UTC)
    assert client.not_before == datetime(2021, 10, 29, 17, 53, 46, tzinfo=UTC)
    assert client.comment == ""
    assert client.copy_extensions is False


@pytest.mark.django_db
def test_with_stdin_with_der() -> None:
    """Test passing a path to a DER."""
    loaded_certificate = x509.load_pem_x509_certificate(CMC_CLIENT_ONE_PEM)
    cmc_client_der = loaded_certificate.public_bytes(Encoding.DER)

    stdout = StringIO()
    with stdin_bytes(cmc_client_der):
        call_command("cmc_add_client", "-", stdout=stdout)
    assert stdout.getvalue() == f"Client with serial {serial} added successfully.\n"

    client = CMCClient.objects.get()
    assert client.certificate.pem.strip() == CMC_CLIENT_ONE_PEM.decode()
    assert client.serial == serial
    assert client.not_after == datetime(2026, 10, 29, 17, 53, 46, tzinfo=UTC)
    assert client.not_before == datetime(2021, 10, 29, 17, 53, 46, tzinfo=UTC)
    assert client.comment == ""


def test_with_path_does_not_exist() -> None:
    """Test passing a path that does not exist."""
    with pytest.raises(CommandError, match=r"^/does/not/exist: File does not exist\.$"):
        call_command("cmc_add_client", "/does/not/exist")


def test_with_invalid_data() -> None:
    """Test passing a path that does not exist."""
    with pytest.raises(CommandError, match=r"^Cannot parse certificate\.$"):
        call_command("cmc_add_client", __file__)



"""CMC management command."""

import os
import sys
from typing import Any

from cryptography import x509
from django.core.management import CommandError, CommandParser
from django_ca.management.base import BaseCommand

from django_ca_cmc.models import CMCClient


class Command(BaseCommand):
    """Command class."""

    help = "Add a CMC client certificate."

    def read_file(self, path: str) -> bytes:
        """Read data from stdin or from a file."""
        if path == "-":
            return sys.stdin.buffer.read()
        if not os.path.exists(path):
            raise CommandError(f"{path}: File does not exist.")
        with open(path, mode="rb") as stream:
            return stream.read()

    def add_arguments(self, parser: CommandParser) -> None:
        """Add arguments."""
        parser.add_argument(
            "-c", "--comment", default="", help="Optional information about this certificate."
        )
        parser.add_argument(
            "--copy-extensions",
            default=False,
            action="store_true",
            help="Copy (almost) all extensions from the CSR if the request is signed by "
            "this certificate.",
        )
        parser.add_argument("certificate", help="Path to the certificate (or - for stdin).")

    def handle(self, certificate: str, comment: str, copy_extensions: bool, **options: Any) -> None:
        """Handle method."""
        raw_certificate = self.read_file(certificate)
        try:
            loaded_certificate = x509.load_pem_x509_certificate(raw_certificate)
        except ValueError:
            try:
                loaded_certificate = x509.load_der_x509_certificate(raw_certificate)
            except ValueError as ex:
                raise CommandError("Cannot parse certificate.") from ex

        client = CMCClient(comment=comment, copy_extensions=copy_extensions)
        client.update_certificate(loaded_certificate)
        client.save()

        self.stdout.write(f"Client with serial {client.serial} added successfully.")



"""django-ca-cmc provides CMC support for django-ca."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("django-ca-cmc")
except PackageNotFoundError:
    # package is not installed
    pass
